#!/etc/opt/openit/perl
#
# $Id: psm.dm,v 1.17 2001/11/01 15:49:53 borgar Exp $
#
# (C) Copyright 2002 by Open iT Norge AS, All rights reserved.
#
# This software may be used or copied only with the written permission of
# Open Systems Computing AS or in accordance with the terms and conditions
# stipulated in the agreement under which this software has been supplied.
#
# This file is part of Open iT Resources.
#
#HEAD##########################################################################
#
# @NAME:
#   patch
# @AUTHOR:
#   Borgar Olsen
# @CREATED:
#   Wed Oct 16 3:47:01 MET 2002
# @DESCRIPTION:
#   Install a patch in a Open iT installation
#
#/HEAD#########################################################################

#MAIN##########################################################################
#
# @HEADING Patch installation
#
# This program, which must be inculded in with all patches, will install
# a patch. This basically consists of replacing files and restarting
# the appropriate daemons
#
# The patching is configured in the file 'patch.cfg' which is included in
# the tar file.
#
#/MAIN#########################################################################

use strict;

my $cfg = "patch.cfg";
my $status = 1;
my $reinstall = 0;

my $patch_id;

umask(0022);

# Only root can patch
if ( $< ne 0 ) {
    die "Error: Patching must be run as 'root'!\n";
}

# Verify that we can find the patch.cfg file.
if (! -e $cfg) {
    die "Error: Unable to find patch config file '$cfg'!\n";
}

# Find version of Open iT installation
my $openit_version = &get_installed_version();

# Find architecture of installation
my $arch = &get_architecture();

# Find patch ID and what Open iT verson valid for
my ($patch, $requires, @versions) = &get_patch_info();

# Verify that patch can be used for current installation
if(! match_version($openit_version, @versions)) {
    die "Error: Patch $patch is not valid for Open iT $openit_version!\n";
}

my $arg;

foreach $arg (@ARGV) {
    if ($arg eq "-reinstall") {
        $reinstall = 1;
    } else {
        print "Usage: ./patch.pl [-reinstall]\n";
        exit 0;
    }
}

# Find list of previously installed patches
my %patches = &get_installed_patches();
print "Installed patches: ", join(", ", sort keys %patches), "\n";

# Verify that patch isn't already installed
if ($patches{$patch}) {
    if ($reinstall) {
        print "Warning: Patch '$patch' is already installed, replacing it.\n";
    } else {
        die "Error: Patch '$patch' is already installed!\n";
    }
}

# Verify that required patches are installed
print "Pre-required patches: $requires\n";
my @requires = split(/,/, $requires);
foreach my $req (@requires) {
    $req =~ s/\s//g;
    if (! $patches{$req}) {
	die "Error: Pre-required patch '$req' has not been installed!\n"
    }
}

# Read file and loop through all lines
open (CFG, "patch.cfg");
while ($status && (my $line = <CFG>)) {
    next if ( $line =~ /^\s*\#.*|^\s*$/ );  # Skip comments and empty lines
    chomp($line);

    # Handle directives in $lines[0]
    $status = &handle_operations(split(":", $line));
}
close(CFG);

# Updated patch info for installation
&update_patch_info($patch);

# Output patch status (i.e ok / failed etc).
if (! $status) {
    die "Error: Patching failed!\n";
} else {
    print "Patch installed!\n";
}


exit 0;

sub get_installed_version {
    my $file = "/var/opt/openit/inst/package/common.info";

    open(IN, $file) || die "Open iT not installed on client!\n";
    my $line = <IN>; # Only need first line
    close(IN);

    my $version = (split(" ", $line))[1];

    # Add missing .0's
    my $dots = $version =~ tr/././;
    for (; $dots < 3 ; $dots++) {
        $version .= ".0";
    }
    print "Patching Open iT $version\n";

    return $version;
}

sub get_patch_info {
    open(INFO, "README") || die "Unable to read README file!\n";
    my @lines = <INFO>;
    close(INFO);

    # Get ID
    my ($id) = grep( s/\+ID:\s*(.*)\s*/$1/, @lines);

    # Get versions
    my ($versions) = grep( s/\+VALID FOR:\s*(.*)\s*/$1/, @lines);
    my @versions = split(",", $versions);

    # Get requires
    my ($requires) = grep( s/\+REQUIRES:\s*(.*\S)\s*/$1/, @lines);

    my @requires = ();
    # REQUIRES may contain a installation version which will
    # superseed the patch requirement. Test for this
    foreach my $element( split( /,/, $requires ) ) {
	if ($element =~ /or/) {
	    my ($patch, $req_version) = (split(/ or /, $element));
	    if ( ! &required_version($req_version) ) {
		push @requires, $patch; #
	    }
	} else {
	    push @requires, $element;
	}
    }
    $requires = join ",", @requires;

    return ($id, $requires, @versions);
}

sub get_architecture {
    my $file = "/var/opt/openit/inst/product/base.prod";

    open(IN, $file) || die "Open iT not installed on client!\n";
    my $line = <IN>; # Only need first line
    close(IN);

    return (split(" ", $line))[0];
}

sub required_version {
    my ($version) = @_;

    my @instparts = split('\.', $openit_version);
    my @reqparts = split('\.', $version);

    # Make sure version number in 4 parts (e.g 4.9.0.0)
    while (scalar(@reqparts) < 4) {
	push @reqparts, 0;
    }

    my $level = 0;
    while ($level < 4) {
	my $ip = $instparts[$level];
	my $rp = $reqparts[$level];

	if ( $ip > $rp ) { # Version higher than requireed (accept)
	    return 1;
	} elsif ( $ip < $rp ) { # Version lower (fail)
	    return 0;
	}
	$level++;
    }

    return 1;
}

sub match_version {
    my ($rel, @vrels) = @_;

    foreach my $vrel (@vrels) {
	# Remove spaces
	$vrel =~ s/\s*(.*)\s*$/$1/;

	# Transform to match
	$vrel =~ s/\./\\./g; # . -> \.
	$vrel =~ s/x/\\d+/g;  # x -> \d+

	if ($rel =~ /^$vrel/) {
	    return 1;
	}
    }

    return 0;
}

sub get_installed_patches {
    my $dir = "/var/opt/openit/inst/patches/";
    my %patches;

    if (! -d $dir) { # If dir dosn't exist, no patche have been applied
	return ();
    }

    opendir(PATCHES, $dir) || "Unable to open patch info dir '$dir'\n";
    my @files = readdir PATCHES;

    closedir(PATCHES);
    @files = grep /^\d+$/, @files;

    foreach $patch (@files) {
	$patches{$patch} = 1,
    }

    return %patches;
}


sub handle_operations {
    my ($op, @params) = @_;

    if ($op eq "replace") {
	&replace_file(@params);
    } elsif($op eq "restart") {
	&restart(@params);
    } elsif($op eq "run") {
	&run_program(@params);
    } elsif($op eq "unpack") {
	&unpack_tar(@params);
    } elsif($op eq "add") {
	&add_file(@params);
    } else {
	print "Error: Unknwon operation '$op' found!\n";
	return 0;
    }

    return 1;
}

sub replace_file {
    my ($type, $source, $dest) = @_;
    print "Replacing file '$source'!\n";

    # Modify source and destination (for binaries)
    if ($type eq "b") {
	$source = "files/$arch/$source";
	$dest =~ s/\$ARCH/$arch/;  # Replace
    } else {
	$source = "files/$source";
    }

    # Verify that source file exists
    if (! -e $source) {
	die "Error:: File '$source' is missing from patch!\n";
    }

    # Verify that destination file exists
    if (! -e $dest) {
	print "Warning: File '$dest' is missing, skipping update!\n";
	return;
    }

    system("mv $dest $dest.old");
    system("cp $source $dest");

}

sub add_file {
    my ($type, $source, $dest) = @_;
    print "Adding file '$source'!\n";

    # Modify source and destination (for binaries)
    if ($type eq "b") {
	$source = "files/$arch/$source";
	$dest =~ s/\$ARCH/$arch/;  # Replace
    } else {
	$source = "files/$source";
    }

    # Verify that source file exists
    if (! -e $source) {
	die "Error:: File '$source' is missing from patch!\n";
    }

    # Verify that destination directory exists
    my $dir = $dest;
    if ( $dest !~ m#^(/|/.*/)[^/]+$# ) {
	print "Warning: Filepath '$dest' is invalid, skipping update!\n";
	return;
    } else {
	$dir = $1;
    }
    if (! -d $dir ) {
	print "Warning: Didn't find expected directory '$dir'," .
	  " skipping update!\n";
	return;
    }

    # Process unexpected old file and adding new
    if( -e $dest ) {
	print "Warning: '$dest' already exists. Renaming!\n";
	system("cp $dest $dest.old");
    }
    system("cp $source $dest");

}

sub restart {
    my ($type) = @_;

    if ($type eq "client") {
	print "Restarting Open iT client\n";
	system("/opt/openit/bin/openit-client restart");
    } elsif ($type eq "server") {
	print "Restarting Open iT server\n";
	system("/opt/openit/bin/openit-server restart");
    } elsif ($type eq "webserver") {
	print "Restarting Open iT web server\n";
	system("/opt/openit/bin/openit-httpd restart");
    } else {
	print "Error: Restart of unknown daemon '$type'!\n";
    }

}

sub run_program {
    my ($prog, @params) = @_;

    print "Running program '$prog'!\n";

    my $command;

    if ($prog =~ /^\//) { # Full path
	$command = "$prog " . join(" ", @params);
    } else {                   # else
	$command = "./files/$prog " . join(" ", @params);
    }

    if( system("$command") ) {
	die "Error: program '$prog' failed!";
    }
}

sub unpack_tar {
    my ($archive, $dir) = @_;

    print "Unpacking archive '$archive'\n";

    # Sanity check (no remove of /);
    if ($dir =~ q'^/$' || $dir eq "") {
	die "Illegal to remove '$dir'";
    }

    # Remove directory if it exists, then create
    if (-d $dir) {
	system("rm -rf $dir");
    }
    mkdir $dir;

    my $command = "/opt/openit/bin/tar xf files/$archive -C $dir";
    if( system("$command") ) {
	die "Unpacking of '$archive' failed!";
    }
    system("chown -R root $dir");
    system("chmod -R a+rX $dir");
}

sub update_patch_info {
    my ($id) = @_;
    my $dir = "/var/opt/openit/inst/patches/";

    if (! -d $dir) { # Create patch dir if missing
	mkdir $dir || die "Error: Unable to create patch dir '$dir'\n";
    }

    # Copy README file to patch dir

    system("cp README $dir/$id");
}
